import functools
import http.server
import logging
import threading

import pytest
import testfixtures


@pytest.fixture(autouse=True)
def cap_log():
    with testfixtures.LogCapture(
        "mkdocs.plugins.mkdocs_section_index",
        attributes=("levelname", "getMessage"),
        ensure_checks_above=logging.WARNING,
    ) as capture:
        yield capture


@pytest.fixture(scope="session")
def http_server(tmp_path_factory):
    directory = tmp_path_factory.mktemp("http_server")
    httpd = http.server.HTTPServer(
        ("localhost", 0),
        functools.partial(http.server.SimpleHTTPRequestHandler, directory=str(directory)),
    )
    t = threading.Thread(target=httpd.serve_forever)
    t.daemon = True
    t.start()
    httpd.directory = directory
    httpd.url = f"http://localhost:{httpd.server_port}/"
    return httpd



import contextlib
import dataclasses
import pathlib

import pytest
from jinja2 import Environment, FileSystemLoader
from mkdocs.config import Config, load_config
from mkdocs.structure.files import File, get_files
from mkdocs.structure.nav import Navigation, get_navigation
from testfixtures import StringComparison

from mkdocs_section_index import SectionPage, plugin

example_dir = pathlib.Path(__file__).parent / ".." / "example"


@pytest.mark.parametrize("directory_urls", ["use_directory_urls", "no_directory_urls"])
@pytest.mark.parametrize("nav", ["explicit_nav", "derived_nav"])
def test_real_example(tmpdir, directory_urls, nav):
    config = dict(
        docs_dir=str(example_dir / "docs"),
        site_dir=tmpdir,
        use_directory_urls=(directory_urls == "use_directory_urls"),
        nav=load_config(str(example_dir / "mkdocs.yml"))["nav"] if nav == "explicit_nav" else None,
    )
    files = get_files(config)
    nav = get_navigation(files, config)
    nav = plugin.SectionIndexPlugin().on_nav(nav, config, files)

    assert len(nav.pages) == 5
    assert len(nav.items) == 3

    assert nav.items[1].is_page
    assert nav.items[1].file.name == "baz"
    assert not nav.items[1].is_section

    sec = nav.items[2]
    assert isinstance(sec, SectionPage)
    assert sec.is_section
    assert sec.is_page
    assert sec.title == "Borgs"
    assert sec.url in {"borgs/", "borgs/index.html"}
    assert sec.file.name == "index"

    assert len(sec.children) == 2
    assert sec.children[0].is_page
    assert sec.children[0].file.name == "bar"

    assert nav.items[1].next_page == sec
    assert sec.children[1].parent == sec


@dataclasses.dataclass
class FakeFiles:
    config: Config

    def get_file_from_path(self, path):
        if ":" not in path:
            return File(
                path,
                src_dir=self.config.get("docs_dir", ""),
                dest_dir=self.config.get("site_dir", ""),
                use_directory_urls=self.config.get("use_directory_urls", True),
            )
        return None

    @classmethod
    def documentation_pages(cls):
        return []


@pytest.mark.golden_test("navs/*.yml")
def test_nav_repr(golden, tmpdir):
    for use_directory_urls in True, False:
        config = dict(nav=golden["input"], use_directory_urls=use_directory_urls)
        files = FakeFiles(config)
        nav = get_navigation(files, config)
        nav = plugin.SectionIndexPlugin().on_nav(nav, config, files)
        assert str(nav) == golden.out[use_directory_urls]


@contextlib.contextmanager
def template_test(directory):
    config = {}
    files = FakeFiles(config)
    env = Environment(loader=FileSystemLoader(directory))  # noqa: S701
    env.filters["url"] = lambda s: s

    plg = plugin.SectionIndexPlugin()
    plg.on_nav(Navigation([], []), config, files)
    env = plg.on_env(env, config, files)

    yield env

    plg.on_post_build(config)


def test_build_material(tmpdir):
    directory = tmpdir.mkdir("material")
    directory.mkdir("partials").join("nav-item.html").write_text(
        "md-nav__icon\n{{ nav_item.title }}", encoding="utf-8"
    )
    with template_test(directory) as env:
        env.get_template("partials/nav-item.html")


def test_build_wrong_content(cap_log, tmpdir):
    directory = tmpdir.mkdir("material")
    directory.mkdir("partials").join("nav-item.html").write_text(
        "href=\n{{ nav_item.title }}", encoding="utf-8"
    )
    with template_test(directory) as env:
        env.get_template("partials/nav-item.html")
    cap_log.check(("WARNING", StringComparison("Failed to adapt.+nav-item.+")))


def test_build_wrong_file(cap_log, tmpdir):
    directory = tmpdir.mkdir("foo")
    directory.mkdir("partials").join("nav-item.html").write_text(
        "{{ nav_item.title }}", encoding="utf-8"
    )
    with template_test(directory) as env:
        env.get_template("partials/nav-item.html")
    cap_log.check(("WARNING", StringComparison(".+couldn't detect a supported theme.+")))



import pytest

from mkdocs_section_index import rewrites


@pytest.mark.golden_test("rewrites/mkdocs-sitemap-*.yml")
def test_rewrite_mkdocs_sitemap(golden):
    assert rewrites._transform_mkdocs_sitemap_template(golden["input"]) == golden.out["output"]


@pytest.mark.golden_test("rewrites/readthedocs-base-*.yml")
def test_rewrite_readthedocs_base(golden):
    assert rewrites._transform_readthedocs_base_template(golden["input"]) == golden.out["output"]


@pytest.mark.golden_test("rewrites/material-nav-item-*.yml")
def test_rewrite_material_nav_item(golden):
    assert rewrites._transform_material_nav_item_template(golden["input"]) == golden.out["output"]


@pytest.mark.golden_test("rewrites/material-tabs-item-*.yml")
def test_rewrite_material_tabs_item(golden):
    assert rewrites._transform_material_tabs_item_template(golden["input"]) == golden.out["output"]



import os
import re
import textwrap
from pathlib import Path

import mechanicalsoup
import pytest
from mkdocs.commands.build import build
from mkdocs.config.base import load_config


class _Browser(mechanicalsoup.StatefulBrowser):
    def __init__(self, url: str):
        super().__init__(soup_config={"features": "html.parser"}, raise_on_404=True)
        self.open(url)

    def links(self, text=None, *args, **kwargs):
        all_links = super().links(*args, **kwargs)
        if text is not None:
            all_links = [a for a in all_links if a.text.strip() == text]
        return all_links


def build_site(cfg: str, src_dir: os.PathLike, dest_dir: os.PathLike) -> None:
    cfg_prefix = f"""
        site_name: SiteName
        site_dir: {str(dest_dir)!r}
        plugins:
            - section-index
    """
    cfg = textwrap.dedent(cfg_prefix) + textwrap.dedent(cfg)

    docs_dir = Path(src_dir, "docs")
    docs_dir.mkdir()
    Path(docs_dir, "README.md").write_text("", encoding="utf-8")

    for rel_path in re.findall(r"\b[^ ]+\.md$", cfg, flags=re.MULTILINE):
        path = Path(docs_dir, rel_path)
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_text(f"@{rel_path}@", encoding="utf-8")

    f = Path(src_dir, "mkdocs.yml")
    f.write_text(cfg, encoding="utf-8")
    build(load_config(str(f)))


@pytest.mark.parametrize("use_directory_urls", [True, False])
@pytest.mark.parametrize(
    ("theme", "features"),
    [
        ("readthedocs", []),
        ("material", []),
        ("material", ["navigation.tabs"]),
        ("material", ["navigation.sections"]),
    ],
)
def test_nav_basic(http_server, tmpdir, use_directory_urls, theme, features):
    cfg = f"""
        use_directory_urls: {use_directory_urls!r}
        theme:
            name: {theme!r}
            features: {features}
        nav:
            - SectionWithIndex1:
                - Sub1/index.md
                - Aaa: Sub1/aaa.md
            - SectionWithIndex2:
                - Sub2/home.md
                - Bbb: Sub1/bbb.md
            - SectionWithoutIndex1:
                - Ccc: Sub2/ccc.md
                - Ddd: Sub2/ddd.md
    """
    build_site(cfg=cfg, src_dir=tmpdir, dest_dir=http_server.directory)
    browser = _Browser(http_server.url)
    # import webbrowser; webbrowser.open(browser.get_url()); input("...")

    browser.follow_link(text="SectionWithIndex1")
    assert browser.get_current_page().find_all(text="@Sub1/index.md@")

    browser.follow_link(text="SectionWithIndex2")
    assert browser.get_current_page().find_all(text="@Sub2/home.md@")

    assert len(browser.links(text="SectionWithIndex2")) >= 1 + features.count("navigation.tabs")
    assert len(browser.links(text="SectionWithoutIndex1")) >= features.count("navigation.tabs")

    browser.follow_link(text="Bbb")
    assert browser.get_current_page().find_all(text="@Sub1/bbb.md@")


@pytest.mark.parametrize("use_directory_urls", [True, False])
@pytest.mark.parametrize(
    ("theme", "features"),
    [
        ("material", []),
        ("material", ["navigation.tabs"]),
        ("material", ["navigation.sections"]),
    ],
)
def test_nav_nested_tabs(http_server, tmpdir, use_directory_urls, theme, features):
    cfg = f"""
        use_directory_urls: {use_directory_urls!r}
        theme:
            name: {theme!r}
            features: {features}
        nav:
            - TabWithoutIndex1:
                - SectionWithIndex1:
                    - Sub1/index.md
                    - Aaa: Sub1/aaa.md
                - SectionWithIndex2:
                    - Sub2/home.md
                    - Bbb: Sub1/bbb.md
            - TabWithoutIndex2:
                - SectionWithoutIndex1:
                    - Ccc: Sub2/ccc.md
                    - Ddd: Sub2/ddd.md
                - SectionWithoutIndex2:
                    - Eee: Sub2/eee.md
    """
    build_site(cfg=cfg, src_dir=tmpdir, dest_dir=http_server.directory)
    browser = _Browser(http_server.url)
    # import webbrowser; webbrowser.open(browser.get_url()); input("...")

    browser.follow_link(text="SectionWithIndex1")
    assert browser.get_current_page().find_all(text="@Sub1/index.md@")
    browser.follow_link(text="SectionWithIndex2")
    assert browser.get_current_page().find_all(text="@Sub2/home.md@")

    if "navigation.tabs" in features:
        browser.follow_link(text="TabWithoutIndex1")
        assert browser.get_current_page().find_all(text="@Sub1/index.md@")
    assert not browser.links(text="SectionWithoutIndex1")


@pytest.mark.parametrize("use_directory_urls", [True, False])
@pytest.mark.parametrize(
    ("theme", "features"),
    [
        ("material", []),
        ("material", ["navigation.tabs"]),
        ("material", ["navigation.sections"]),
    ],
)
def test_nav_nested_tabs_2(http_server, tmpdir, use_directory_urls, theme, features):
    cfg = f"""
        use_directory_urls: {use_directory_urls!r}
        theme:
            name: {theme!r}
            features: {features}
        nav:
            - TabWithIndex1:
                - Sub1/index.md
                - SectionWithoutIndex1:
                    - Aaa: Sub1/aaa.md
    """
    build_site(cfg=cfg, src_dir=tmpdir, dest_dir=http_server.directory)
    browser = _Browser(http_server.url)
    # import webbrowser; webbrowser.open(browser.get_url()); input("...")

    browser.follow_link(text="TabWithIndex1")
    assert browser.get_current_page().find_all(text="@Sub1/index.md@")



from __future__ import annotations

import itertools
import logging
import pathlib
import textwrap
from collections.abc import Callable

from jinja2 import BaseLoader, Environment

__all__ = ["TemplateRewritingLoader"]


log = logging.getLogger(f"mkdocs.plugins.{__name__}")


class TemplateRewritingLoader(BaseLoader):
    def __init__(self, loader: BaseLoader):
        self.loader = loader
        self.found_supported_theme = False

    def get_source(
        self, environment: Environment, template: str
    ) -> tuple[str, str, Callable[[], bool] | None]:
        src: str | None
        src, filename, uptodate = self.loader.get_source(environment, template)
        old_src = src
        assert filename is not None
        path = pathlib.Path(filename).as_posix()

        if path.endswith("/mkdocs/templates/sitemap.xml"):
            src = _transform_mkdocs_sitemap_template(src)
        else:
            # the second path is used in MkDocs-Material >= 9.4
            if path.endswith(
                (
                    "/material/partials/nav-item.html",
                    "/material/templates/partials/nav-item.html",
                ),
            ):
                src = _transform_material_nav_item_template(src)
            elif path.endswith(
                (
                    "/material/partials/tabs-item.html",
                    "/material/templates/partials/tabs-item.html",
                ),
            ):
                src = _transform_material_tabs_item_template(src)
            elif path.endswith("/themes/readthedocs/base.html"):
                src = _transform_readthedocs_base_template(src)
            elif path.endswith(("/properdocs_theme_readthedocs/base.html", "/nature/base.html")):
                src = None  # Just works!
            else:
                return src, filename, uptodate
            self.found_supported_theme = True

        if old_src == src:
            log.warning(
                f"Failed to adapt the theme file '{filename}'. "
                f"This is likely a bug in mkdocs-section-index, and things won't work as expected."
            )
        return src or old_src, filename, uptodate


def _transform_mkdocs_sitemap_template(src: str) -> str | None:
    if " in pages " in src:
        return None
    # The below only for versions <= 1.1.2.
    return src.replace(
        "{%- else %}",
        "{%- endif %}{% if item.url %}",
    )


def _transform_material_nav_item_template(src: str) -> str:
    if "navigation.indexes" in src:
        return (
            src.replace(  # For versions >= 9.6.10
                "_ = namespace(index = none)",
                "_ = namespace(index = nav_item if nav_item.url else none)",
            )
            .replace(  # For versions < 9.6.10
                "{% set indexes = [] %}",
                "{% set indexes = [nav_item] if nav_item.url else [] %}",
            )
            .replace(
                "{% if nav_item.children | length > 1 %}",
                "{% if nav_item.children %}",
            )
        )

    # The above only for versions >= 7.3, the below only for versions < 7.3.
    src = src.replace(
        "{% if nav_item.children %}",
        "{% if nav_item.children and not ('navigation.tabs' in features and level == 1 and not nav_item.active and nav_item.url) %}",
    )

    repl = """\
        {% if nav_item.url %}
          <a href="{{ nav_item.url | url }}" class="md-nav__link{% if nav_item == page %} md-nav__link--active{% endif %}"
            style="margin: initial; padding: initial; pointer-events: initial">
        {% endif %}
          [...]
        {% if nav_item.url %}</a>{% endif %}
    """
    lines = src.split("\n")
    for i, (line1, line2) in enumerate(itertools.pairwise(lines)):
        for a, b in (line1, line2), (line2, line1):
            if "md-nav__icon" in a and b.endswith("{{ nav_item.title }}"):
                lines[i : i + 2] = (a, _replace_line(b, repl))
                break
    return "\n".join(lines)


def _transform_material_tabs_item_template(src: str) -> str:
    src = src.replace(
        "{% if first.children %}",
        "{% if first.children and not first.url %}",
    )
    # The above only for versions >= 9.2
    src = src.replace(
        "{% if nav_item.children %}",
        "{% if nav_item.children and not nav_item.url %}",
    )
    # The above only for versions > 6.1.7, the below only for versions <= 6.1.7.
    return src.replace(
        "(nav_item.children | first).url",
        "(nav_item.url or (nav_item.children | first).url)",
    ).replace(
        "if (nav_item.children | first).children",
        "if (nav_item.children | first).children and not nav_item.url",
    )


def _transform_readthedocs_base_template(src: str) -> str | None:
    if " if nav_item.is_page " in src:
        return None
    # The below only for versions < 1.6:
    repl = """\
        {% if nav_item.url %}
            <ul><li{% if nav_item == page %} class="current"{% endif %}>
                <a href="{{ nav_item.url|url }}" style="padding: 0; font-size: inherit; line-height: inherit">
        {% endif %}
                [...]
        {% if nav_item.url %}
                </a>
            </li></ul>
        {% endif %}
    """
    lines = src.split("\n")
    for i, line in enumerate(lines):
        if "{{ nav_item.title }}" in line:
            lines[i] = _replace_line(lines[i], repl)
    return "\n".join(lines)


def _replace_line(line: str, wrapper: str, new_line: str | None = None) -> str:
    leading_space = line[: -len(line.lstrip())]
    if new_line is None:
        new_line = line.lstrip()
    new_text = textwrap.dedent(wrapper.rstrip()).replace("[...]", new_line)
    return textwrap.indent(new_text, leading_space)



from __future__ import annotations

import collections
import logging
from typing import TYPE_CHECKING

from mkdocs.plugins import BasePlugin
from mkdocs.structure.nav import Navigation, Section
from mkdocs.structure.pages import Page

from . import SectionPage, rewrites

if TYPE_CHECKING:
    from jinja2 import Environment

try:
    import properdocs.replacement_warning

    # Warn when this plugin is being used from the mkdocs executable.
    properdocs.replacement_warning.setup()
except ImportError:
    pass

__all__ = ["SectionIndexPlugin"]

log = logging.getLogger(f"mkdocs.plugins.{__name__}")


class SectionIndexPlugin(BasePlugin):
    def on_nav(self, nav: Navigation, config, files) -> Navigation:
        todo = collections.deque((nav.items,))
        while todo:
            items = todo.popleft()
            for i, section in enumerate(items):
                if not isinstance(section, Section) or not section.children:
                    continue
                todo.append(section.children)
                page = section.children[0]
                if not isinstance(page, Page):
                    continue
                assert not page.children
                if not page.title and page.url:
                    # The page becomes a section-page.
                    page.__class__ = SectionPage
                    assert isinstance(page, SectionPage)
                    page.is_section = page.is_page = True
                    page.title = section.title
                    page.parent = section.parent
                    # The page leaves the section but takes over children that used to be its peers.
                    section.children.pop(0)
                    page.children = section.children
                    for child in page.children:
                        child.parent = page
                    # The page replaces the section; the section will be garbage-collected.
                    items[i] = page
        self._nav = nav
        return nav

    def on_env(self, env: Environment, config, files) -> Environment:
        assert env.loader is not None
        env.loader = self._loader = rewrites.TemplateRewritingLoader(env.loader)
        return env

    def on_page_context(self, context, page, config, nav):
        if nav != self._nav:
            self._nav = nav
            log.warning(
                "It seems that the effects of section-index plugin have been lost, because another MkDocs plugin re-wrote the nav! "
                "Re-order `plugins` in mkdocs.yml so that 'section-index' appears closer to the end."
            )

    def on_post_build(self, config):
        if not self._loader.found_supported_theme:
            log.warning(
                "section-index plugin couldn't detect a supported theme to adapt. "
                "It probably won't work as expected. "
                "See https://github.com/oprypin/mkdocs-section-index#theme-support"
            )



from __future__ import annotations

from mkdocs.structure.nav import Section
from mkdocs.structure.pages import Page

__version__ = "0.3.12"
__all__ = ["SectionPage"]


class SectionPage(Section, Page):  # type: ignore[misc]
    def __init__(self, title: str, file, config, children):
        Page.__init__(self, title=title, file=file, config=config)
        Section.__init__(self, title=title, children=children)
        self.is_section = self.is_page = True

    active = Page.active  # type: ignore[assignment]

    def __repr__(self):
        result = Page.__repr__(self)
        if not result.startswith("Section"):
            result = "Section" + result
        return result

    def __eq__(self, other):
        return object.__eq__(self, other)

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return object.__hash__(self)
